package org.codehaus.groovy.grails.plugins.gwt

import java.beans.Introspector
import java.lang.annotation.Annotation
import java.lang.reflect.Method
import java.lang.reflect.Modifier
import java.lang.reflect.ParameterizedType 
import java.lang.reflect.Type 
import java.util.Map;
import java.util.regex.Matcher 

import org.apache.commons.logging.Log
import org.apache.commons.logging.LogFactory
import org.codehaus.groovy.grails.plugins.gwt.GwtServiceInterfaceGenerator
import org.codehaus.groovy.grails.plugins.gwt.annotation.CollectionTypeArg
import org.codehaus.groovy.grails.plugins.gwt.annotation.MapTypeArg
import java.beans.PropertyDescriptor
import org.codehaus.groovy.grails.commons.GrailsClassUtils

import com.google.gwt.user.client.rpc.RemoteServiceRelativePath;

/**
 * Created by IntelliJ IDEA.
 * User: pal20
 * Date: 02-Apr-2009
 * Time: 09:00:14
 * To change this template use File | Settings | File Templates.
 */
class BetterGwtServiceInterfaceGenerator implements GwtServiceInterfaceGenerator {
    String srcPath = "src/java";

    boolean isGwtExposed(Class serviceClass) {
        return getPackage(serviceClass) != null
    }

    boolean getInterfacesExist(Class serviceClass) {
        return getInterfacesExist(serviceClass, getPackage(serviceClass))
    }

    boolean getInterfacesExist(Class serviceClass, String packageName) {
        File outputDir = getOutputDir(packageName)
        return new File(outputDir, getMainFilename(serviceClass)).exists() &&
                new File(outputDir, getAsyncFilename(serviceClass)).exists()
    }

    /**
     * Creates the required service and async interfaces for a given
     * Grails service. This checks to make sure that neither interface
     * file has been modified since they were last generated. If either
     * has, then the generation is skipped - we don't want to overwrite
     * user changes.
     * @param serviceWrapper (GrailsClass) The Grails service's class.
     * @param packageName (String) The GWT client package in which to
     * put the interfaces.
     * @param log The plugin logger.
     */
    void generateInterfaces(Class serviceClass) {
        String pkg = getPackage(serviceClass)

        // Ignore this service class if the package is null.
        if (pkg != null) {
            generateInterfaces(serviceClass, pkg)
        }
    }

    void generateInterfaces(Class serviceClass, String packageName) {
        // Find the directory in which to store the interface files.
        def outputDir = new File(srcPath, packageName.replace('.' as char, '/' as char))

        // Make sure that the output directory exists so that we can
        // create the interface files in it.
        if (!outputDir.exists()) {
            // The directory doesn't exist - it should really, since
            // a GWT module should reside here. Give the user the
            // benefit of the doubt and attempt to create the directory,
            // but log a warning in case the package is incorrect.
            log.warn "Directory '${outputDir}' does not exist - creating it now."
            if (!outputDir.mkdirs()) {
                log.error "Could not create required output directory."
                return
            }
        }

        // The generated interface files.
        def className = serviceClass.simpleName
        def mainFile = new File(outputDir, "${className}.java")
        def asyncFile = new File(outputDir, "${className}Async.java")

        // Start the content of the main interface definition.
        def output = new StringBuilder(1024)
        def outputAsync = new StringBuilder(1024)
        writeInterfaces(serviceClass, packageName, output, outputAsync)

        // Write the definitions to the appropriate files.
        mainFile.write(output.toString())
        asyncFile.write(outputAsync.toString())
    }

    /**
     * Determines whether the given service has been configured for
     * GWT RPC, and if so returns the package in which the corresponding
     * interfaces should be generated. If the package is specified in
     * the expose list using the "gwt:<package>" notation, then that
     * package is returned. Otherwise, this returns the package of the
     * given service class.
     * @param serviceClass The Grails service to query.
     * @return The fully-qualified name of the package in which to put
     * the generated interfaces, or <code>null</code> if the expose list
     * does not contain an entry for GWT.
     */
    String getPackage(Class serviceClass) {
        def exposeList = GrailsClassUtils.getStaticPropertyValue(serviceClass, 'expose')

        // Check whether 'gwt' is in the expose list.
        def gwtExposed = exposeList?.find { it.startsWith('gwt') }
        if (gwtExposed) {
            def m = gwtExposed =~ 'gwt:(.*)'
            if (m) {
                return m[0][1]
            }
            else {
                return getDefaultPackage(serviceClass)
            }
        }
        else {
            return null
        }
    }
	
	boolean isServiceInterfaceMethod(Method method, Set propMethods) {
		return !method.synthetic && Modifier.isPublic(method.modifiers) && !Modifier.isStatic(method.modifiers) && !propMethods.contains(method)
	}

    protected void writeInterfaces(Class serviceClass, String packageName, mainOutput, asyncOutput) {
		ServiceInterfaces serviceInterfaces = new ServiceInterfaces(packageName:packageName, grailsServiceSimpleClass:serviceClass.simpleName)

		def propMethods = getPropertyMethods(serviceClass)
		
        // Iterate through the methods declared by the Grails service,
        // adding the appropriate ones to the interface definitions.
		def declaredMethods = serviceClass.declaredMethods.sort { it.name }
        declaredMethods.each {Method method ->
            // Skip non-public, static, Groovy, and property methods.
            if (!isServiceInterfaceMethod(method, propMethods)) {
                return
            }
			serviceInterfaces.addMethod(TypeUtil.getReturnTypeAsString(method), method.name, TypeUtil.getParameterTypesAsString(method), TypeUtil.getThrowTypesAsString(method))
        }
		String syncIfc = serviceInterfaces.toSyncInterface()
		String asyncIfc = serviceInterfaces.toAsyncInterface()

        mainOutput << syncIfc
        asyncOutput << asyncIfc
    }

	protected Set getPropertyMethods(Class serviceClass) {
		// Find out what properties the service class contains, because
		// we want to leave them out of the interface definition.
		def info = Introspector.getBeanInfo(serviceClass)
		def propMethods = [] as Set
		info.propertyDescriptors.each {PropertyDescriptor desc ->
			propMethods << desc.readMethod
			propMethods << desc.writeMethod

			// Groovy adds a "get*()" method for booleans as well as
			// the usual "is*()", so we have to remove it too.
			if (desc.readMethod?.name?.startsWith("is")) {
				def name = "get${desc.readMethod.name[2..-1]}".toString()
				def getMethod = info.methodDescriptors.find { it.name == name }
				if (getMethod) {
					propMethods << getMethod.method
				}
			}
		}
		return propMethods
	}
	
    protected File getOutputDir(String packageName) {
        // Find the directory in which to store the interface files.
        return new File(srcPath, packageName.replace('.' as char, '/' as char))
    }

    protected String getMainFilename(Class serviceClass) {
        return "${serviceClass.simpleName}.java"
    }

    protected String getAsyncFilename(Class serviceClass) {
        return "${serviceClass.simpleName}Async.java"
    }

    protected String getDefaultPackage(Class serviceClass) {
        Package pkg = serviceClass.getPackage()
        return pkg ? pkg.name + ".client" : "client"
    }

    private Log getLog() {
        return LogFactory.getLog(this.getClass())
    }
}


class ServiceInterfaces {
	String packageName
	String grailsServiceSimpleClass
	
	ImportStatements importStatements = new ImportStatements()
	
	List<ServiceMethod> methods = []
	
	
	void addMethod(String returnType, String name, List parameterTypes, List throwTypes) {
		methods << new ServiceMethod(importStatements, returnType, name, parameterTypes, throwTypes)
	}
	
	String toSyncInterface() {
		return """\
package ${packageName};

${importStatements.toSyncImports()}

@RemoteServiceRelativePath("rpc")
public interface ${grailsServiceSimpleClass} extends RemoteService {
    ${methods*.toSyncMethod().join('\n    ')}
}
"""
			
	}  
	
	String toAsyncInterface() {
		return """\
package ${packageName};

${importStatements.toAsyncImports()}

public interface ${grailsServiceSimpleClass}Async {
    ${methods*.toAsyncMethod().join('\n    ')}
}
"""
	}
}


class ImportStatements {
	Map<String, String> map = [:]
	
	String addComplexType(String complexType) {
		if (complexType ==~ /(void|boolean[\[\]]*|byte[\[\]]*|short[\[\]]*|int[\[\]]*|long[\[\]]*|float[\[\]]*|double[\[\]]*|char[\[\]]*)/) {
			return complexType
		}
		List<String> allRawTypes = TypeUtil.getAllRawTypes(complexType)
		List importedTypes = [] 
		for (String nextType : allRawTypes) {
			if (add(nextType)) {
				importedTypes << nextType
			}
		}
		importedTypes.each {
			complexType = complexType.replace(it, TypeUtil.toSimpleTypeName(it))
		}
		complexType = TypeUtil.eliminateJavaLang(complexType);
		return complexType
	}
	private boolean add(String qualifiedType) {
		qualifiedType = qualifiedType.replace("[]", "")
		if (qualifiedType && !map.containsKey(qualifiedType) && !qualifiedType.startsWith("java.lang.")) {
			String simpleType = TypeUtil.toSimpleTypeName(qualifiedType)
			if (!map.containsValue(simpleType)) {
				map.put(qualifiedType, simpleType)
				return true
			}
		}
		return map.containsKey(qualifiedType)
	}
	
	int size() {
		return map.size()
	}

	String toSyncImports() {
		List<String> allImports = []
		map.each { key, value ->
			allImports << "import ${key};"
		}
		allImports << "import com.google.gwt.user.client.rpc.RemoteService;"
		allImports << "import com.google.gwt.user.client.rpc.RemoteServiceRelativePath; "
		return allImports.sort().join("\n")
	}
	
	String toAsyncImports() {
		List<String> allImports = []
		map.each { key, value ->
			allImports << "import ${key};"
		}
		allImports << "import com.google.gwt.user.client.rpc.AsyncCallback;"
		return allImports.sort().join("\n")
	}
}



class ServiceMethod {
	ImportStatements importStatements
	
	String returnType
	String name
	List<ServiceMethodParameter> parameters = []
	List<String> throwTypes = []

	ServiceMethod(ImportStatements importStatements, String returnType, String name, List parameterTypes, List throwTypes) {
		this.importStatements = importStatements
		setReturnType(returnType)
		this.name = name
		if (parameterTypes) {
			for (int i in 0..<parameterTypes.size()) {
				addParameter(parameterTypes[i], "arg$i")
			}
		}
		if (throwTypes) {
			for (int i in 0..<throwTypes.size()) {
				addThrowType(throwTypes[i])
			}
		}
	}
	
	void setReturnType(String returnType) {
		this.returnType = importStatements.addComplexType(returnType)
	}
	
	void addParameter(String type, String name) {
		parameters << new ServiceMethodParameter(importStatements, type, name)
	}
	
	void addThrowType(String throwType) {
		throwTypes << importStatements.addComplexType(throwType)
	}
	
	String toSyncMethod() {
		String parameterString = parameters*.toString().join(', ')
		String throwsClause = throwTypes ? " throws ${throwTypes.join(', ')}" : ""
		return "public $returnType $name(${parameterString})${throwsClause};"
	} 

	String toAsyncMethod() {
		String parameterString = parameters*.toString().join(', ')
		String cbSep = parameters ? ", " : ""
		String cbType = (returnType == "void") ? "Void" : returnType
		String cb = "AsyncCallback<$cbType> callback"
		return "public void $name(${parameterString}${cbSep}${cb});"
	} 
}

class ServiceMethodParameter {
	ImportStatements importStatements
	
	String type
	String name
	
	ServiceMethodParameter(ImportStatements importStatements, String type, String name) {
		this.type = importStatements.addComplexType(type);
		this.name = name
	}
	
	String toString() {
		return "$type $name"
	}
}

class TypeUtil {
	static String getReturnTypeAsString(Method method) {
		return getTypes(method.genericReturnType)[0]
	}

	static List<String> getParameterTypesAsString(Method method) {
		return getTypes(method.genericParameterTypes)
	}
	
	static List<String> getThrowTypesAsString(Method method) {
		return getTypes(method.genericExceptionTypes)
	}
	static List<String> getTypes(Type[] genericTypes) {
		List typesAsString = []

		for (int i in 0..<genericTypes.size()) {
			def nextGenericType = getParametrizedType(genericTypes[i])
			if (nextGenericType) {
				typesAsString << "${nextGenericType}"
			} else {
				typesAsString << "${getType(genericTypes[i])}"
			}
		}
		return typesAsString

	}
	static ParameterizedType getParametrizedType(Type returnType) {
		if (returnType instanceof ParameterizedType) {
			ParameterizedType pt = ((ParameterizedType)returnType)
			def actual = pt.actualTypeArguments
			if (actual) {
				return pt
			}
		}
	}
	
	/**
	 * Returns the string representation of the given type. For example,
	 * 'java.lang.String', 'int', 'java.lang.String[]', 'boolean[][][]'.
	 */
	static String getType(Class clazz) {
		if (!clazz.array) {
			// If the type is not an array, we can simply return its
			// name.
			return clazz.name
		}
		else {
			// The class name contains some number of '[' characters
			// indicating the dimensions of the array.
			def dimensions = clazz.name.count('[')

			// To get the base type of the array, we have to recurse
			// through the component types.
			def type = clazz.componentType
			for (int i in 1..<dimensions) {
				type = type.componentType
			}

			return type.name + '[]' * dimensions
		}
	}

	
	
	static List<String> getAllRawTypes(String complexType) {
		if (complexType ==~ /[^<>,\s]*/) {
			return [complexType]
		}
		return complexType.split(/[<>, ]/) as List
	}
	static String eliminateJavaLang(String complexType) {
		return complexType.replace("java.lang.", "")
	}
	static String eliminateArrayBrackets(String complexType) {
		return complexType.replace("[]", "")
	}
	static String toSimpleTypeName(String qualifiedTypeName) {
		Matcher m = qualifiedTypeName =~/.*\.([^.]+)/
		if (m) {
			return m[0][1]
		}
		return qualifiedTypeName
	}
}